package com.sillycycle.bagleyd.ant3d;

/* Code for selection spliced in is due to Alex Shvets */
/* http://home.comcast.net/~shvets/Programs/Examples/GameLife/life2.html */
// Ant3DMatrix.java

import java.io.Externalizable;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.sillycycle.bagleyd.util.Box;
import com.sillycycle.bagleyd.util.CellInfo;

public class Ant3DMatrix implements Externalizable, Comparator<CellInfo> {
	private static final long serialVersionUID = 42L;

	private int columns, rows, stacks;
	private int area, volume;

	private byte[] cells;
	public transient List<CellInfo> list;

	// must exist for deserialization process
	public Ant3DMatrix() {
	}

	public Ant3DMatrix(int columns, int rows, int stacks) {
		this.columns = columns;
		this.rows = rows;
		this.stacks = stacks;
		area = columns * rows;
		volume = area * stacks;
		list = new LinkedList<>();
		init();
	}

	@Override
	public int compare(CellInfo a, CellInfo b) {
		if (a.getDistance() < b.getDistance())
			return 1;
		if (a.getDistance() > b.getDistance())
			return -1;
		return 0;
	}

	public final void init() {
		cells = new byte[volume];
		clear();
	}

	public void clear() {
		for (int position = 0; position < volume; position++)
			cells[position] = (byte) 0; // Empty
	}

	public void clearLiveCells() {
		for (Iterator<CellInfo> iterator =
				list.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int position = cellInfo.getPosition();
			cells[position] = 0;
			iterator.remove();
		}
	}

	// O(n)
	public void clearCell(int position) {
		for (Iterator<CellInfo> iterator =
				list.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int p = cellInfo.getPosition();
			if (position == p)
				iterator.remove();
			//break; // commented to clear duplications
		}
		cells[position] = 0;
	}

	public int getRows() {
		return rows;
	}

	public int getColumns() {
		return columns;
	}

	public int getStacks() {
		return stacks;
	}

	public byte getCell(int position) {
		byte value = cells[position];

		if (value == -128)
			return 0;
		else if (value < 0)
			return (byte) -value;
		else
			return value;
	}

	public boolean getCellBit(int position) {
		return (cells[position] < 0);
	}

	public void setArrayCell(int position, byte cell) {
		cells[position] = cell;
	}

	// O(n)
	public void setListCell(int position, byte cell) {
		for (CellInfo cellInfo : list) {
			int p = cellInfo.getPosition();
//System.out.println("setCell: " + p + ",cell " +cell);
if (position == p && cell != 0) {
	cellInfo.setCell(cell);
	return;
}
		}
		list.add(new CellInfo(position, cell));
	}


	public void setCell(int position, byte cell) {
		setListCell(position, cell);
		setArrayCell(position, cell);
	}

	public void setCellBit(int position) {
		byte value = cells[position];

		if (value == 0)
			value = -128;
		else if (value > 0)
			value = (byte) -value;
		cells[position] = value;
	}

	public void resetCellBit(int position) {
		byte value = cells[position];

		if (value == -128)
			value = 0;
		else if (value < 0)
			value = (byte) -value;
		cells[position] = value;
	}

	public byte getCell(int column, int row, int stack) {
		if ((stack >= 0) && (stack < stacks) &&
				(row >= 0) && (row < rows) &&
				(column >= 0) && (column < columns))
			return getCell(stack * area + row * columns + column);
		return 0;
	}

	// To be used so do not add to dead list for cells already added
	public boolean getCellBit(int column, int row, int stack) {
		if ((stack >= 0) && (stack < stacks) &&
				(row >= 0) && (row < rows) &&
				(column >= 0) && (column < columns))
			return getCellBit(stack * area + row * columns + column);
		return true;
	}

	public void setCell(int column, int row, int stack, byte cell) {
		if ((stack >= 0) && (stack < stacks) &&
				(row >= 0) && (row < rows) &&
				(column >= 0) && ( column < columns)) {
			int position = stack * area + row * columns + column;
			if (cell == 0) {
				clearCell(position);
			} else {
				setCell(position, cell);
			}

		}
	}
	/*public void setCell(int column, int row, int stack, byte cell) {
		if ((stack >= 0) && (stack < stacks) &&
				(row >= 0) && (row < rows) &&
				(column >= 0) && ( column < columns))
			cells[stack * area + row * columns + column] = cell;
	}*/

	public Ant3DMatrix getSubMatrix(Box b) {
		Ant3DMatrix selection = new Ant3DMatrix(b.width, b.height, b.depth);

		for (int k = 0; k < b.depth; k++)
		for (int j = 0; j < b.height; j++)
		for (int i = 0; i < b.width; i++) {
			selection.setCell(i, j, k,
				getCell(i + b.x, j + b.y, k + b.z));
		}
		return selection;
	}

	public void setSubMatrix(Ant3DMatrix selection, Box b) {
		for (int k = 0; k < b.depth; k++)
		for (int j = 0; j < b.height; j++)
		for (int i = 0; i < b.width; i++) {
			try {
				setCell(i + b.x, j + b.y, k + b.z,
					selection.getCell(i, j, k));
			} catch (ArrayIndexOutOfBoundsException e) {
				// ignore
			}
		}
	}

	public void clearSubMatrix(Box b) {
		for (int k = b.z; k < b.z + b.depth; k++)
		for (int j = b.y; j < b.y + b.height; j++)
		for (int i = b.x; i < b.x + b.width; i++) {
			cells[k * area + j * columns + i] = 0; //Empty
		}
	}

	/* The center point of rectangle (more or less) is used
	 * to rotate about.
	 */
	void rotate(Ant3DMatrix selection, Box b,
			int polyhedron, int plane, boolean cw) {
		for (int k = 0; k < b.depth; k++)
		for (int j = 0; j < b.height; j++)
		for (int i = 0; i < b.width; i++) {
			if (polyhedron == 6) {
				if (plane == 0) { /* z */
					if (cw)
						setCell(b.height - j - 1 + b.x,
							i + b.y,
							k + b.z,
							selection.getCell(i, j, k));
					else
						setCell(j + b.x,
							b.width - i - 1 + b.y,
							k + b.z,
							selection.getCell(i, j, k));
				}
				if (plane == 1) { /* x */
					if (cw)
						setCell(i + b.x,
							b.depth - k - 1 + b.y,
							j + b.z,
							selection.getCell(i, j, k));
					else
						setCell(i + b.x,
							k + b.y,
							b.height - j - 1 + b.y,
							selection.getCell(i, j, k));
				}
			}
		}
	}

	/* Wall means you reflecting in a wall mirror,
	 * else it means you are reflecting off ground.
	 */
	void reflect(Ant3DMatrix selection, Box b,
			int polyhedron, int plane, boolean wall) {
		for (int k = 0; k < b.depth; k++)
		for (int j = 0; j < b.height; j++)
		for (int i = 0; i < b.width; i++) {
			if (polyhedron == 6) {
				if (plane == 0) {
					if (wall)
						setCell(b.width - i - 1 + b.x,
							j + b.y,
							k + b.z,
							selection.getCell(i, j, k));
					else
						setCell(i + b.x,
							b.height - j - 1 + b.y,
							k + b.z,
							selection.getCell(i, j, k));
				}
			}
		}
	}


	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();

		for (int k = 0; k < stacks; k++) {
		for (int j = 0; j < rows; j++) {
			for (int i = 0; i < columns; i++)
				sb.append(cells[k * area + j * columns + i]);
			sb.append("\n");
		}
		sb.append("\n");
		}
		sb.append("\n");
		return sb.toString();
	}


	// implements Externalizable

	@Override
	public void writeExternal(ObjectOutput out) throws IOException {
		out.writeByte(columns);
		out.writeByte(rows);
		out.writeByte(stacks);
		for (int k = 0; k < stacks; k++)
		for (int j = 0; j < rows; j++)
		for (int i = 0; i < columns; i++) {
			if (cells[k * area + j * columns + i] == 1) {
				out.writeByte(i);
				out.writeByte(j);
				out.writeByte(k);
			}
		}
	}

	@Override
	public void readExternal(ObjectInput in)
			throws IOException, ClassNotFoundException {
		columns = in.readByte();
		rows = in.readByte();
		stacks = in.readByte();
		init();
		while (true) {
			try {
				int i = in.readByte();
				int j = in.readByte();
				int k = in.readByte();
				cells[k * area + j * columns + i] = 1;
			} catch (IOException e)	{
				break;
			}
		}
	}

	public static void main(String[] args) {
		System.out.println("Constructing object:");
		Ant3DMatrix m = new Ant3DMatrix(9, 8, 7);
		m.setCell(2, 3, 2, (byte) 1);
		m.setCell(3, 3, 2, (byte) 1);
		m.setCell(4, 3, 2, (byte) 1);
		m.setCell(4, 4, 2, (byte) 1);
		m.setCell(3, 5, 2, (byte) 1);
		m.setCell(2, 3, 3, (byte) 1);
		m.setCell(3, 3, 3, (byte) 1);
		m.setCell(4, 3, 3, (byte) 1);
		m.setCell(4, 4, 3, (byte) 1);
		m.setCell(3, 5, 3, (byte) 1);
		Ant3DMatrix n;

		System.out.println("Saving object:");
		try (ObjectOutputStream out = new ObjectOutputStream(
				new FileOutputStream("junk.out"))) {
			out.writeObject(m);
		} catch (Exception e) {
			e.printStackTrace();
		}
		// Now get it back
		System.out.println("Recovering object:");
		try (ObjectInputStream in = new ObjectInputStream(
				new FileInputStream("junk.out"))) {
			n = (Ant3DMatrix)in.readObject();
			System.out.println(n.toString());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
